"""
    MrTopo Error - identifies possible errors & handles output
"""

from enum import Enum
from mrtopo.logger import log

# Enum for error identities
class ErrorIdentities(Enum):
    UNDEFINED = -1
    BAD_JSON_INPUT = 1
    INVALID_CONFIG = 2
    

# Error handling method
def handle_error(e_type=ErrorIdentities.UNDEFINED, message=""):
    o_msg = "MrTopo Error - " + str(e_type.name) + "\n" + str(message)
    log(item=o_msg ,level=e_type.value)


"""
    MrTopo - Mutator - handles mutation of networks
"""
from mrtopo.logger import log
from mrtopo.structures.mutantnetwork import MutantNetwork
from mrtopo.mutator.operators import Operations, do
from shutil import copyfile
from math import floor
import random

GENERATIONS = 30


def mutate(network):
    log("Mutator - mutating network " + str(GENERATIONS) + " times")
    mutant_networks = []  # type MutantNetwork
    for i in range(GENERATIONS):
        operation = random.choice(list(Operations))
        mn = do(operation, network.deep_copy(), i)  # mutate deep copy of network
        if mn:
            mutant_networks.append(mn)
    log(mutant_networks)
    return mutant_networks


def get_var_names(coll):
    names = []
    for item in coll:
        name = ""
        for c in item[0]:
            if c == "=":
                break
            else:
                name += str(c)
        names.append(name.strip())
    return names


def mutated_lines(n_remove, network_arr):
    deleted = []  # links to be removed

    for i in range(n_remove):
        deleted.append(random.choice(network_arr))

    return deleted



"""
    MrTopo Tester - tests MrTopo mutated topologies
"""
import os
import shutil
import subprocess
from mrtopo.logger import log


def test(dir, target_file, command_file):

    mutated_file_paths = list_py_files(dir)

    made = False

    while not made:
        try:
            os.mkdir("MrTopoTest")
            made = True
        except OSError:
            log("MrTopoTest exists, removing")
            shutil.rmtree("MrTopoTest")

    f = open(f"MrTopoTest/original_topology.py", "x")
    f.close()
    shutil.copyfile(target_file, "MrTopoTest/original_topology.py")

    test_out = []

    for path in mutated_file_paths:
        out = log(f"Executing test with: {path}") + "\n"
        shutil.copyfile(path, target_file)
        out += call(f"sh {command_file}")
        test_out.append(out)

    shutil.copyfile("MrTopoTest/original_topology.py", target_file)
    return test_out


def list_py_files(dest):
    _dir = os.listdir(dest)
    py_file_paths = []
    for _file in _dir:
        n = len(_file)
        if _file[n - 2] + _file[n - 1] == "py":
            if dest[len(dest)-1] == "/":
                dest = dest[:-1]
            py_file_paths.append(f'{dest}/{_file}')
    return py_file_paths


def call(command):
    try:
        return log(subprocess.check_output(command, shell=True).decode('utf-8'))
    except subprocess.CalledProcessError:
        return log(f'Command: {command} - failed')


from mrtopo.structures.network import Network


class MutantNetwork():
    def __init__(self, network: Network, _id: int, description: str, operation):
        self.network = network
        self._id = _id
        self.description = description
        self.operation = operation
        self.modified_item = None

    def __repr__(self):
        return "<MutantNetwork (" + str(self.network) + ", " + str(self._id) + ", " + str(
            self.description) + ", " + str(self.operation) + ", Modified - " + str(self.modified_item) + ")>"

    def get_id(self):
        return self._id

    def get_network(self):
        return self.network

    def get_description(self):
        return self.description

    def get_operation(self):
        return self.operation

    def get_modified_item(self):
        return self.modified_item

    def add_modified_item(self, mi):
        self.modified_item = mi



"""
    MrTopo - translator.mrtopoio - io handling
"""

import json, glob, re
from mrtopo.logger import log
from mrtopo.util.iofunctor import vald_def


# gets and validates input based on validator output - default returns True on all input
# on None invalid
def get_input(message: str, validator=vald_def):
    u_in = input(message)

    log('Validating ~' + u_in)

    if validator(u_in):
        log('\n~' + u_in + ' valid')
        return u_in

    log('Invalid input ~' + u_in + ' try again...')
    return None


# Function to read files of the type json
def read_json(file: str):
    _json = None
    with open(file) as jfile:
        _json = json.load(jfile)
    return _json


# looks for file in current directory
def find_file(file: str):
    return glob.glob('./' + file)


def open_read(file: str):
    return open(file, "r")



"""
    MrTopo - util.iofunctor - functors for validation of user input
"""


# default validator
def vald_def(u_in):
    return True


# valid yes or no
def vald_y_n(u_in):
    if (u_in == 'y' or u_in == 'n'):
        return True
    return False



"""
    MrTopo Interpreter - interprets translated input to mutable data structures
"""

from mrtopo.logger import log
from mrtopo.structures import config, network


def interpret(cfg: config.Config):
    log('Interpreter - interpreting config')
    return c_construct_networks(cfg.topologies)


def interpret(py: dict):
    log('Interpreter - interpreting py file')
    return p_construct_networks(py)


def p_construct_networks(py: dict):
    return network.Network.p_build(py)


def c_construct_networks(topologies: list):
    networks = []
    for t in topologies:
        networks.append(network.Network.c_build(t))
    return networks



"""
    MrTopo Validator - validates MrTopo mutated topologies
"""

import os
import subprocess

from mrtopo.logger import log


def validate(file, name=None, long=False):
    log(f"Validating: {file}")

    if not name:
        name = extract_name(file)

    log(f"Topology name: {name}")

    construct(file, name)

    test(file, name, long)

    if long:
        res_f = ['dmp', 'bch', 'pga']
    else:
        res_f = ['dmp', 'bch']

    results = []

    for f in res_f:
        try:
            results.append(f'\t* Command: {find_cmd(f, file, name)}\n>>> <{open(f, "r").read()}>')
        except:
            results.append(f'\t* Command: {find_cmd(f, file, name)} - failed')

    log(f"Success!")

    descriptor = "Details:"

    for res in results:
        if res != "":
            descriptor += f"\n{res}\n"

    del_all()

    return bundle(file, descriptor)


def construct(file, name):
    if os.geteuid() != 0:
        log(f"Root access is required to continue the execution of this script.")
    res = call(f'sudo mn --custom {file} --topo {name} --test dump > dmp')
    return res


def test(file, name, long):
    benchmark = f"Benchmark:\n{call(f'sudo mn --custom {file} --topo {name} --test none > bch')}"
    if long:
        pingall = f"Pingall:\n{call(f'sudo mn --custom {file} --topo {name} --test pingall > pga')}"
        return [benchmark, pingall]
    else:
        return [benchmark]


def extract_name(file):
    with open(file, 'r') as f:
        _f = f.readlines()

    for line in _f:
        if line.__contains__('topos') and line.__contains__('{'):
            if line.__contains__("\'"):
                return line.split("\'")[1]
            elif line.__contains__('\"'):
                return line.split('\"')


def bundle(file, descriptor):
    return f"Bundle {file}:\n{descriptor}\n"


def call(command):
    try:
        return log(subprocess.check_output(command, shell=True).decode('utf-8'))
    except subprocess.CalledProcessError:
        return log(f'Command: {command} - failed')


def del_all():
    all = ['bch', 'pga', 'dmp']
    for f in all:
        try:
            os.remove(f)
        except FileNotFoundError:
            log(f'{f} does not exist')


def find_cmd(fn, file, name):
    if fn == 'bch':
        return f'sudo mn --custom {file} --topo {name} --test none > bch'
    elif fn == 'pga':
        return f'sudo mn --custom {file} --topo {name} --test pingall > pga'
    elif fn == 'dmp':
        return f'sudo mn --custom {file} --topo {name} --test dump > dmp'